@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_timestamp.c,v 1.6 2010/02/01 09:05:20 t-ishii Exp $
+ * $Header$
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2009	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -25,14 +25,17 @@
 
 #include "pool.h"
 #include "pool_timestamp.h"
+#include "pool_relcache.h"
+#include "pool_select_walker.h"
+#include "pool_config.h"
 #include "parser/parsenodes.h"
-#include "parser/gramparse.h"
+#include "parser/parser.h"
 #include "parser/pool_memory.h"
 
-#define Assert(x)
 
 typedef struct {
-	char	*attrname;
+	char	*attrname;	/* attribute name */
+	char	*adsrc;		/* default value expression */
 	int		 use_timestamp;
 } TSAttr;
 
@@ -61,6 +64,7 @@ static bool rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx);
 static bool rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx);
 static char *get_current_timestamp(POOL_CONNECTION_POOL *backend);
 static Node *makeTsExpr(TSRewriteContext *ctx);
+static A_Const *makeStringConstFromQuery(POOL_CONNECTION_POOL *backend, char *expression);
 bool raw_expression_tree_walker(Node *node, bool (*walker) (), void *context);
 
 #define		MAX_RELCACHE 32
@@ -70,6 +74,9 @@ POOL_RELCACHE	*ts_relcache;
 static void *
 ts_register_func(POOL_SELECT_RESULT *res)
 {
+/* Number of result columns included in res */
+#define NUM_COLS		3
+
 	TSRel	*rel;
 	int		 i;
 
@@ -80,10 +87,22 @@ ts_register_func(POOL_SELECT_RESULT *res)
 
 	for (i = 0; i < res->numrows; i++)
 	{
-		rel->attr[i].attrname = strdup(res->data[i * 2]);
-		rel->attr[i].use_timestamp = *(res->data[i * 2 + 1]) == 't';
-		pool_debug("attrname %s use_timestamp = %d",
-			rel->attr[i].attrname, rel->attr[i].use_timestamp);
+		int index = 0;
+
+		rel->attr[i].attrname = strdup(res->data[i * NUM_COLS + index]);
+		index++;
+
+		if (res->data[i * NUM_COLS + index])
+			rel->attr[i].adsrc = strdup(res->data[i * NUM_COLS + index]);
+		else
+			rel->attr[i].adsrc = NULL;
+
+		index++;
+
+		rel->attr[i].use_timestamp = *(res->data[i * NUM_COLS + index]) == 't';
+		pool_debug("attrname %s adsrc %s use_timestamp = %d",
+				   rel->attr[i].attrname, (rel->attr[i].adsrc? rel->attr[i].adsrc:"NULL"),
+										   rel->attr[i].use_timestamp);
 	}
 
 	rel->relnatts = res->numrows;
@@ -107,15 +126,44 @@ ts_unregister_func(void *data)
 static TSRel*
 relcache_lookup(TSRewriteContext *ctx)
 {
-#define ATTRDEFQUERY "SELECT attname, coalesce(d.adsrc = 'now()' OR d.adsrc LIKE '%%''now''::text%%', false)" \
+#define ATTRDEFQUERY "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
+	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
+	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
+	" a.atttypid = 'date'::regtype::oid OR" \
+	" a.atttypid = 'time'::regtype::oid OR" \
+	" a.atttypid = 'time with time zone'::regtype::oid)" \
+    " , false)" \
 	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
 	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
 	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s'" \
 	" ORDER BY a.attnum"
 
+#define ATTRDEFQUERY2 "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
+	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
+	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
+	" a.atttypid = 'date'::regtype::oid OR" \
+	" a.atttypid = 'time'::regtype::oid OR" \
+	" a.atttypid = 'time with time zone'::regtype::oid)" \
+    " , false)" \
+	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
+	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
+	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = pgpool_regclass('%s')" \
+	" ORDER BY a.attnum"
+
+	char *query;
+
+	if (pool_has_pgpool_regclass())
+	{
+		query = ATTRDEFQUERY2;
+	}
+	else
+	{
+		query = ATTRDEFQUERY;
+	}
+
 	if (!ts_relcache)
 	{
-		ts_relcache = pool_create_relcache(MAX_RELCACHE, ATTRDEFQUERY, ts_register_func, ts_unregister_func, false);
+		ts_relcache = pool_create_relcache(MAX_RELCACHE, query, ts_register_func, ts_unregister_func, false);
 
 		if (ts_relcache == NULL)
 		{
@@ -189,7 +237,7 @@ isSystemTypeCast(Node *node, const char *name)
 		return false;
 
 	typecast = (TypeCast *) node;
-	return isSystemType((Node *) typecast->typename, name);
+	return isSystemType((Node *) typecast->typeName, name);
 }
 
 /*
@@ -208,12 +256,15 @@ rewrite_timestamp_walker(Node *node, void *context)
 		/* `now()' FuncCall */
 		FuncCall	*fcall = (FuncCall *) node;
 
-		if (list_length(fcall->funcname) == 1 &&
-			strcmp("now", strVal(linitial(fcall->funcname))) == 0)
+		if ((list_length(fcall->funcname) == 1 &&
+			 strcmp("now", strVal(linitial(fcall->funcname))) == 0) ||
+			(list_length(fcall->funcname) == 2 &&
+			 strcmp("pg_catalog", strVal(linitial(fcall->funcname))) == 0 &&
+			 strcmp("now", strVal(lsecond(fcall->funcname))) == 0))
 		{
 			TypeCast	*tc = makeNode(TypeCast);
 			tc->arg = makeTsExpr(ctx);
-			tc->typename = SystemTypeName("text");
+			tc->typeName = SystemTypeName("text");
 
 			fcall->funcname = SystemFuncName("timestamptz");
 			fcall->args = list_make1(tc);
@@ -225,11 +276,11 @@ rewrite_timestamp_walker(Node *node, void *context)
 		/* CURRENT_DATE, CURRENT_TIME, LOCALTIMESTAMP, LOCALTIME etc.*/
 		TypeCast	*tc = (TypeCast *) node;
 
-		if ((isSystemType((Node *) tc->typename, "date") ||
-			 isSystemType((Node *) tc->typename, "timestamp") ||
-			 isSystemType((Node *) tc->typename, "timestamptz") ||
-			 isSystemType((Node *) tc->typename, "time") ||
-			 isSystemType((Node *) tc->typename, "timetz")))
+		if ((isSystemType((Node *) tc->typeName, "date") ||
+			 isSystemType((Node *) tc->typeName, "timestamp") ||
+			 isSystemType((Node *) tc->typeName, "timestamptz") ||
+			 isSystemType((Node *) tc->typeName, "time") ||
+			 isSystemType((Node *) tc->typeName, "timetz")))
 		{
 			/* rewrite `'now'::timestamp' and `'now'::text::timestamp' both */
 			if (isSystemTypeCast(tc->arg, "text"))
@@ -316,7 +367,11 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			if (relcache->attr[i].use_timestamp)
 			{
 				rewrite = true;
-				values = lappend(values, makeTsExpr(ctx));
+				if (ctx->rewrite_to_params)
+					values = lappend(values, makeTsExpr(ctx));
+				else
+					values = lappend(values,
+									 makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc));
 			}
 			else
 				values = lappend(values, makeNode(SetToDefault));
@@ -379,7 +434,10 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					if (relcache->attr[i].use_timestamp == true && IsA(lfirst(lc_val), SetToDefault))
 					{
 						rewrite = true;
-						lfirst(lc_val) = makeTsExpr(ctx);
+						if (ctx->rewrite_to_params)
+							lfirst(lc_val) = makeTsExpr(ctx);
+						else
+							lfirst(lc_val) = makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc);
 					}
 					i++;
 				}
@@ -390,7 +448,11 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					if (relcache->attr[i].use_timestamp == true)
 					{
 						rewrite = true;
-						values = lappend(values, makeTsExpr(ctx));
+						if (ctx->rewrite_to_params)
+							values = lappend(values, makeTsExpr(ctx));
+						else
+							values = lappend(values,
+											 makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc));
 					}
 					else
 						values = lappend(values, makeNode(SetToDefault));
@@ -402,12 +464,15 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			/*
 			 * INSERT INTO rel(col1, col2) VALUES (val, val2)
 			 *
-			 * if timestamp column does not given by column list
+			 * if timestamp column is not given by column list
 			 * add colname to column list and add timestamp to values list.
 			 */
 			int			append_columns = 0;
+			int			*append_columns_list;
 			ResTarget	*col;
 
+			append_columns_list = (int *)malloc(sizeof(int)*relcache->relnatts);
+
 			for (i = 0; i < relcache->relnatts; i++)
 			{
 				if (relcache->attr[i].use_timestamp == false)
@@ -430,7 +495,7 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					col->indirection = NIL;
 					col->val = NULL;
 					i_stmt->cols = lappend(i_stmt->cols, col);
-					append_columns++;
+					append_columns_list[append_columns++] = i;
 				}
 			}
 
@@ -451,16 +516,24 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					if (relcache->attr[i].use_timestamp == true && IsA(lfirst(lc_val), SetToDefault))
 					{
 						rewrite = true;
-						lfirst(lc_val) = makeTsExpr(ctx);
+						if (ctx->rewrite_to_params)
+							lfirst(lc_val) = makeTsExpr(ctx);
+						else
+							lfirst(lc_val) = makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc);
 					}
 				}
 
 				/* add ts_const to values list */
 				for (i = 0; i < append_columns; i++)
 				{
-					values = lappend(values, makeTsExpr(ctx));
+					if (ctx->rewrite_to_params)
+						values = lappend(values, makeTsExpr(ctx));
+					else
+						values = lappend(values,
+										 makeStringConstFromQuery(ctx->backend, relcache->attr[append_columns_list[i]].adsrc));
 				}
 			}
+			free(append_columns_list);
 		}
 	}
 
@@ -511,7 +584,10 @@ rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx)
 				{
 					if (relcache->attr[i].use_timestamp)
 					{
-						res->val = (Node *) makeTsExpr(ctx);
+						if (ctx->rewrite_to_params)
+							res->val = (Node *) makeTsExpr(ctx);
+						else
+							res->val = (Node *)makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc);
 						rewrite = true;
 					}
 					break;
@@ -528,7 +604,7 @@ rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx)
  */
 char *
 rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
-	   	bool rewrite_to_params, Portal *portal)
+				  bool rewrite_to_params, POOL_SENT_MESSAGE *message)
 {
 	TSRewriteContext	ctx;
 	Node			*stmt;
@@ -598,11 +674,11 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 		rewrite = ctx.rewrite;
 
 		/* add params */
-		if (portal)
+		if (message)
 		{
 			int		i;
 
-			for (i = 0; i < portal->num_tsparams; i++)
+			for (i = 0; i < message->num_tsparams; i++)
 			{
 				e_stmt->params = lappend(e_stmt->params, ctx.ts_const);
 				rewrite = true;
@@ -615,7 +691,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 	if (!rewrite)
 		return NULL;
 
-	if (ctx.rewrite_to_params && portal)
+	if (ctx.rewrite_to_params && message)
 	{
 		ListCell	*lc;
 		int			 num = ctx.num_params + 1;
@@ -628,7 +704,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 		}
 
 		/* save to portal */
-		portal->num_tsparams = list_length(ctx.params);
+		message->num_tsparams = list_length(ctx.params);
 
 		/* add param type */
 		if (IsA(node, PrepareStmt))
@@ -636,7 +712,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 			int				 i;
 			PrepareStmt		*p_stmt = (PrepareStmt *) node;
 
-			for (i = 0; i < portal->num_tsparams; i++)
+			for (i = 0; i < message->num_tsparams; i++)
 				p_stmt->argtypes =
 				   	lappend(p_stmt->argtypes, SystemTypeName("timestamptz"));
 		}
@@ -662,8 +738,9 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
  * rewrite Bind message to add parameter
  */
 char *
-bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend, Portal *portal,
-		const char *orig_msg, int *len)
+bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
+					   POOL_SENT_MESSAGE *message,
+					   const char *orig_msg, int *len)
 {
 	int16		 tmp2,
 				 num_params,
@@ -677,6 +754,14 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend, Portal *portal,
 				*copy_to,
 				*new_msg;
 
+#ifdef TIMESTAMPDEBUG
+	fprintf(stderr, "message length:%d\n", *len);
+	for(i=0;i<*len;i++)
+	{
+		fprintf(stderr, "%02x ", orig_msg[i]);
+	}
+#endif
+
 	ts = get_current_timestamp(backend);
 	if (ts == NULL)
 	{
@@ -686,8 +771,8 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend, Portal *portal,
 
 	ts_len = strlen(ts);
 
-	*len += (strlen(ts) + sizeof(int32)) * portal->num_tsparams;
-	new_msg = copy_to = (char *) malloc(*len + portal->num_tsparams * sizeof(int16));
+	*len += (strlen(ts) + sizeof(int32)) * message->num_tsparams;
+	new_msg = copy_to = (char *) malloc(*len + message->num_tsparams * sizeof(int16));
 	copy_from = orig_msg;
 
 	/* portal_name */
@@ -708,30 +793,30 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend, Portal *portal,
 	if (num_formats > 1)
 	{
 		/* enlarge message length */
-		*len += portal->num_tsparams * sizeof(int16);
-		tmp2 += portal->num_tsparams;
+		*len += message->num_tsparams * sizeof(int16);
+		tmp2 += message->num_tsparams;
 	}
 	tmp2 = htons(tmp2);
-	memcpy(copy_to, &tmp2, copy_len);
+	memcpy(copy_to, &tmp2, copy_len);	/* copy number of format codes */
 	copy_to += copy_len; copy_from += copy_len;
 
 	copy_len = num_formats * sizeof(int16);
 
-	memcpy(copy_to, copy_from, copy_len);
+	memcpy(copy_to, copy_from, copy_len);		/* copy format codes */
 	copy_to += copy_len; copy_from += copy_len;
 
 	if (num_formats > 1)
 	{
 		/* set format codes to zero(text) */
-		memset(copy_to, 0, portal->num_tsparams * 2);
-		copy_to += sizeof(int16) * portal->num_tsparams;
+		memset(copy_to, 0, message->num_tsparams * 2);
+		copy_to += sizeof(int16) * message->num_tsparams;
 	}
 
 	/* num params */
 	memcpy(&tmp2, copy_from, sizeof(int16));
 	copy_len = sizeof(int16);
 	num_params = ntohs(tmp2);
-	tmp2 = htons(num_params + portal->num_tsparams);
+	tmp2 = htons(num_params + message->num_tsparams);
 	memcpy(copy_to, &tmp2, sizeof(int16));
 	copy_to += copy_len; copy_from += copy_len;
 
@@ -740,14 +825,22 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend, Portal *portal,
 	for (i = 0; i < num_params; i++)
 	{
 		memcpy(&tmp4, copy_from + copy_len, sizeof(int32));
+		tmp4 = ntohl(tmp4);		/* param length */
 		copy_len += sizeof(int32);
-		copy_len += ntohl(tmp4);
+
+		/* If param length is -1, it indicates that the value is NULL
+		 * and we don't have value slot. So we don't add up copy_len.
+		 */
+		if (tmp4 > 0)
+		{
+			copy_len += tmp4;
+		}
 	}
 	memcpy(copy_to, copy_from, copy_len);
 	copy_to += copy_len; copy_from += copy_len;
 
 	tmp4 = htonl(ts_len);
-	for (i = 0; i < portal->num_tsparams; i++)
+	for (i = 0; i < message->num_tsparams; i++)
 	{
 		memcpy(copy_to, &tmp4, sizeof(int32));
 		copy_to += sizeof(int32);
@@ -764,6 +857,39 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend, Portal *portal,
 	return new_msg;
 }
 
+static A_Const *makeStringConstFromQuery(POOL_CONNECTION_POOL *backend, char *expression)
+{
+	A_Const *con;
+	POOL_SELECT_RESULT *res;
+	POOL_STATUS		 status;
+	char query[1024];
+	int len;
+	char *str;
+
+	snprintf(query, sizeof(query), "SELECT %s", expression);
+	status = do_query(MASTER(backend), query, &res, MAJOR(backend));
+	if (status != POOL_CONTINUE)
+	{
+		pool_error("makeStringConstFromQuery: do_query faild");
+		return NULL;
+	}
+
+	if (res->numrows != 1)
+	{
+		free_select_result(res);
+		return NULL;
+	}
+
+	len = strlen(res->data[0]) + 1;
+	str = palloc(len);
+	strcpy(str, res->data[0]);
+	free_select_result(res);
+
+	con = makeNode(A_Const);
+	con->val.type = T_String;
+	con->val.val.str = str;
+	return con;
+}
 
 /* from nodeFuncs.c start */
 
@@ -1002,7 +1128,7 @@ bool
 
 				if (walker(tc->arg, context))
 					return true;
-				if (walker(tc->typename, context))
+				if (walker(tc->typeName, context))
 					return true;
 			}
 			break;
@@ -1053,7 +1179,7 @@ bool
 			{
 				ColumnDef  *coldef = (ColumnDef *) node;
 
-				if (walker(coldef->typename, context))
+				if (walker(coldef->typeName, context))
 					return true;
 				if (walker(coldef->raw_default, context))
 					return true;
@@ -1068,7 +1194,7 @@ bool
 
 				if (walker(xs->expr, context))
 					return true;
-				if (walker(xs->typename, context))
+				if (walker(xs->typeName, context))
 					return true;
 			}
 			break;
